@@ -34,6 +34,20 @@ package api
  *  In this example, a [[scala.reflect.api.Types#TypeRef]] is returned, which corresponds to the type constructor `List`
  *  applied to the type argument `Int`.
  *
+ *  In the case of a generic type, you can also combine it with other types
+ *  using [[scala.reflect.api.Types#appliedType]]. For example:
+ *
+ *  {{{
+ *    scala> val intType = typeOf[Int]
+ *    intType: reflect.runtime.universe.Type = Int
+ *
+ *    scala> val listType = typeOf[List[_]]
+ *    listType: reflect.runtime.universe.Type = List[_]
+ *
+ *    scala> appliedType(listType.typeConstructor, intType)
+ *    res0: reflect.runtime.universe.Type = List[Int]
+ *  }}}
+ *
  *  ''Note:'' Method `typeOf` does not work for types with type parameters, such as `typeOf[List[A]]` where `A` is
  *  a type parameter. In this case, use [[scala.reflect.api.TypeTags#weakTypeOf]] instead.
  *
@@ -51,7 +65,7 @@ package api
  *  For example, to look up the `map` method of `List`, one can do:
  *
  *  {{{
- *     scala> typeOf[List[_]].member("map": TermName)
+ *     scala> typeOf[List[_]].member(TermName("map"))
  *     res1: reflect.runtime.universe.Symbol = method map
  *  }}}
  *
@@ -60,7 +74,6 @@ package api
  *  @groupname TypeCreators Types - Creation
  *  @groupname TypeOps      Types - Operations
  *  @group ReflectionAPI
- *
  *  @contentDiagram hideNodes "*Api"
  */
 trait Types {
@@ -387,7 +400,7 @@ trait Types {
     def map(f: Type => Type): Type
 
     /** Apply `f` to each part of this type, for side effects only */
-    def foreach(f: Type => Unit)
+    def foreach(f: Type => Unit): Unit
 
     /** Returns optionally first type (in a preorder traversal) which satisfies predicate `p`,
      *  or None if none exists.
@@ -497,7 +510,7 @@ trait Types {
     def sym: Symbol
   }
   /** The `SuperType` type is not directly written, but arises when `C.super` is used
-   *  as a prefix in a `TypeRef` or `SingleType`. It's internal presentation is
+   *  as a prefix in a `TypeRef` or `SingleType`. Its internal presentation is
    *  {{{
    *     SuperType(thistpe, supertpe)
    *  }}}
@@ -514,7 +527,7 @@ trait Types {
    */
   val SuperType: SuperTypeExtractor
 
-  /** An extractor class to create and pattern match with syntax `SingleType(thistpe, supertpe)`
+  /** An extractor class to create and pattern match with syntax `SuperType(thistpe, supertpe)`
    *  @group Extractors
    */
   abstract class SuperTypeExtractor {
@@ -540,12 +553,17 @@ trait Types {
      */
     def supertpe: Type
   }
-  /** The `ConstantType` type is not directly written in user programs, but arises as the type of a constant.
-   *  The REPL expresses constant types like `Int(11)`. Here are some constants with their types:
+
+  /** A `ConstantType` type cannot be expressed in user programs; it is inferred as the type of a constant.
+   *  Here are some constants with their types and the internal string representation:
    *  {{{
-   *     1           ConstantType(Constant(1))
-   *     "abc"       ConstantType(Constant("abc"))
+   *     1           ConstantType(Constant(1))       Int(1)
+   *     "abc"       ConstantType(Constant("abc"))   String("abc")
    *  }}}
+   *
+   *  ConstantTypes denote values that may safely be constant folded during type checking.
+   *  The `deconst` operation returns the equivalent type that will not be constant folded.
+   *
    *  @template
    *  @group Types
    */
@@ -1031,7 +1049,21 @@ trait Types {
    */
   def glb(ts: List[Type]): Type
 
-  /** A creator for type applications
+  /** A creator for type applications.
+   *
+   *  Useful to combine and create types out of generic ones. For example:
+   *
+   *  {{{
+   *    scala> val boolType = typeOf[Boolean]
+   *    boolType: reflect.runtime.universe.Type = Boolean
+   *
+   *    scala> val optionType = typeOf[Option[_]]
+   *    optionType: reflect.runtime.universe.Type = Option[_]
+   *
+   *    scala> appliedType(optionType.typeConstructor, boolType)
+   *    res0: reflect.runtime.universe.Type = Option[Boolean]
+   *  }}}
+   *
    *  @group TypeOps
    */
   def appliedType(tycon: Type, args: List[Type]): Type
