@@ -14,12 +14,13 @@ package scala
 package reflect
 package internal
 
-import scala.annotation.elidable
+import java.net.URLClassLoader
+
+import scala.annotation.{elidable, tailrec}
 import scala.collection.mutable
 import util._
 import java.util.concurrent.TimeUnit
 
-import scala.collection.mutable.ArrayBuffer
 import scala.reflect.internal.settings.MutableSettings
 import scala.reflect.internal.{TreeGen => InternalTreeGen}
 import scala.reflect.io.AbstractFile
@@ -73,11 +74,16 @@ abstract class SymbolTable extends macros.Universe
 
   def log(msg: => AnyRef): Unit
 
-  protected def elapsedMessage(msg: String, start: Long) =
-    msg + " in " + (TimeUnit.NANOSECONDS.toMillis(System.nanoTime()) - start) + "ms"
+  protected def elapsedMessage(msg: String, startNs: Long) =
+    msg + " in " + (TimeUnit.NANOSECONDS.toMillis(System.nanoTime() - startNs)) + "ms"
 
-  def informProgress(msg: String)          = if (settings.verbose) inform("[" + msg + "]")
-  def informTime(msg: String, start: Long) = informProgress(elapsedMessage(msg, start))
+  def informProgress(msg: String)            = if (settings.verbose) inform("[" + msg + "]")
+  def informTime(msg: String, startNs: Long) = informProgress(elapsedMessage(msg, startNs))
+  @inline final def informingProgress[T](msg: => String)(fn: => T) : T = {
+    val verbose:Boolean = settings.verbose
+    val start = if(verbose) System.nanoTime() else 0L
+    try fn finally if (verbose) informTime(msg, start)
+  }
 
   def shouldLogAtThisPhase = false
   def isPastTyper = false
@@ -88,12 +94,6 @@ abstract class SymbolTable extends macros.Universe
 
   def settings: MutableSettings
 
-  @deprecated("Interactive is implemented with a custom Global; this flag is ignored", "2.11.0") def forInteractive = false
-  @deprecated("Scaladoc is implemented with a custom Global; this flag is ignored", "2.11.0")    def forScaladoc = false
-
-  @deprecated("use devWarning if this is really a warning; otherwise use log", "2.11.0")
-  def debugwarn(msg: => String): Unit = devWarning(msg)
-
   /** Override with final implementation for inlining. */
   def debuglog(msg:  => String): Unit = if (settings.debug) log(msg)
 
@@ -151,32 +151,19 @@ abstract class SymbolTable extends macros.Universe
     result
   }
 
-  // Getting in front of Predef's asserts to supplement with more info; see `supplementErrorMessage`.
-  // This has the happy side effect of masking the one argument forms of assert/require
-  // (but for now they're reproduced here, because there are a million uses internal and external to fix).
   @inline
-  final def assert(assertion: Boolean, message: => Any): Unit = {
-    // calling Predef.assert would send a freshly allocated closure wrapping the one received as argument.
+  final def assert(assertion: Boolean, message: => Any): Unit =
     if (!assertion) throwAssertionError(message)
-  }
 
-  // Let's consider re-deprecating this in the 2.13 series, to encourage informative messages.
-  //@deprecated("prefer to use the two-argument form", since = "2.12.5")
-  final def assert(assertion: Boolean): Unit = {
-    assert(assertion, "")
-  }
+  @deprecated("consider supplying an explanatory message", since = "2.12.5")
+  final def assert(assertion: Boolean): Unit = assert(assertion, "")
 
   @inline
-  final def require(requirement: Boolean, message: => Any): Unit = {
-    // calling Predef.require would send a freshly allocated closure wrapping the one received as argument.
+  final def require(requirement: Boolean, message: => Any): Unit =
     if (!requirement) throwRequirementError(message)
-  }
 
-  // Let's consider re-deprecating this in the 2.13 series, to encourage informative messages.
-  //@deprecated("prefer to use the two-argument form", since = "2.12.5")
-  final def require(requirement: Boolean): Unit = {
-    require(requirement, "")
-  }
+  @deprecated("consider supplying an explanatory message", since = "2.12.5")
+  final def require(requirement: Boolean): Unit = require(requirement, "")
 
   // extracted from `assert`/`require` to make them as small (and inlineable) as possible
   private[internal] def throwAssertionError(msg: Any): Nothing =
@@ -184,8 +171,8 @@ abstract class SymbolTable extends macros.Universe
   private[internal] def throwRequirementError(msg: Any): Nothing =
     throw new java.lang.IllegalArgumentException(s"requirement failed: ${supplementErrorMessage(String valueOf msg)}")
 
-  @inline final def findSymbol(xs: TraversableOnce[Symbol])(p: Symbol => Boolean): Symbol = {
-    xs find p getOrElse NoSymbol
+  @inline final def findSymbol(xs: IterableOnce[Symbol])(p: Symbol => Boolean): Symbol = {
+    xs.iterator find p getOrElse NoSymbol
   }
 
   // For too long have we suffered in order to sort NAMES.
@@ -203,7 +190,7 @@ abstract class SymbolTable extends macros.Universe
 
   /** Dump each symbol to stdout after shutdown.
    */
-  final val traceSymbolActivity = sys.props contains "scalac.debug.syms"
+  final val traceSymbolActivity = System.getProperty("scalac.debug.syms") != null
   object traceSymbols extends {
     val global: SymbolTable.this.type = SymbolTable.this
   } with util.TraceSymbolActivity
@@ -213,7 +200,7 @@ abstract class SymbolTable extends macros.Universe
   /** Check that the executing thread is the compiler thread. No-op here,
    *  overridden in interactive.Global. */
   @elidable(elidable.WARNING)
-  def assertCorrectThread() {}
+  def assertCorrectThread(): Unit = {}
 
   /** A last effort if symbol in a select <owner>.<name> is not found.
    *  This is overridden by the reflection compiler to make up a package
@@ -235,7 +222,7 @@ abstract class SymbolTable extends macros.Universe
   type RunId = Int
   final val NoRunId = 0
 
-  private val phStack: collection.mutable.ArrayStack[Phase] = new collection.mutable.ArrayStack()
+  private[this] val phStack: collection.mutable.Stack[Phase] = new collection.mutable.Stack()
   private[this] var ph: Phase = NoPhase
   private[this] var per = NoPeriod
 
@@ -246,10 +233,10 @@ abstract class SymbolTable extends macros.Universe
 
   def atPhaseStackMessage = atPhaseStack match {
     case Nil    => ""
-    case ps     => ps.reverseMap("->" + _).mkString("(", " ", ")")
+    case ps     => ps.reverseIterator.map("->" + _).mkString("(", " ", ")")
   }
 
-  final def phase_=(p: Phase) {
+  final def phase_=(p: Phase): Unit = {
     ph = p
     per = period(currentRunId, p.id)
   }
@@ -261,7 +248,7 @@ abstract class SymbolTable extends macros.Universe
     }
     current
   }
-  final def popPhase(ph: Phase) {
+  final def popPhase(ph: Phase): Unit = {
     if (keepPhaseStack) {
       phStack.pop()
     }
@@ -333,23 +320,24 @@ abstract class SymbolTable extends macros.Universe
   final def isValid(period: Period): Boolean =
     period != 0 && runId(period) == currentRunId && {
       val pid = phaseId(period)
-      if (phase.id > pid) infoTransformers.nextFrom(pid).pid >= phase.id
-      else infoTransformers.nextFrom(phase.id).pid >= pid
+      if (phase.id > pid) nextFrom(pid).pid >= phase.id
+      else nextFrom(phase.id).pid >= pid
     }
 
   final def isValidForBaseClasses(period: Period): Boolean = {
+    @tailrec
     def noChangeInBaseClasses(it: InfoTransformer, limit: Phase#Id): Boolean = (
       it.pid >= limit ||
       !it.changesBaseClasses && noChangeInBaseClasses(it.next, limit)
     )
     period != 0 && runId(period) == currentRunId && {
       val pid = phaseId(period)
-      if (phase.id > pid) noChangeInBaseClasses(infoTransformers.nextFrom(pid), phase.id)
-      else noChangeInBaseClasses(infoTransformers.nextFrom(phase.id), pid)
+      if (phase.id > pid) noChangeInBaseClasses(nextFrom(pid), phase.id)
+      else noChangeInBaseClasses(nextFrom(phase.id), pid)
     }
   }
 
-  def openPackageModule(container: Symbol, dest: Symbol) {
+  def openPackageModule(container: Symbol, dest: Symbol): Unit = {
     // unlink existing symbols in the package
     for (member <- container.info.decls.iterator) {
       if (!member.isPrivate && !member.isConstructor) {
@@ -366,7 +354,7 @@ abstract class SymbolTable extends macros.Universe
       }
     }
     // enter decls of parent classes
-    for (p <- container.parentSymbols) {
+    for (p <- container.parentSymbolsIterator) {
       if (p != definitions.ObjectClass) {
         openPackageModule(p, dest)
       }
@@ -398,7 +386,7 @@ abstract class SymbolTable extends macros.Universe
   }
 
   /** if there's a `package` member object in `pkgClass`, enter its members into it. */
-  def openPackageModule(pkgClass: Symbol) {
+  def openPackageModule(pkgClass: Symbol): Unit = {
 
     val pkgModule = pkgClass.packageObject
     def fromSource = pkgModule.rawInfo match {
@@ -411,13 +399,13 @@ abstract class SymbolTable extends macros.Universe
   }
 
   object perRunCaches {
-    import scala.collection.generic.Clearable
+    import scala.collection.mutable.Clearable
 
     // Weak references so the garbage collector will take care of
     // letting us know when a cache is really out of commission.
     import java.lang.ref.WeakReference
-    private var caches = List[WeakReference[Clearable]]()
-    private var javaCaches = List[JavaClearable[_]]()
+    private[this] var caches = List[WeakReference[Clearable]]()
+    private[this] var javaCaches = List[JavaClearable[_]]()
 
     def recordCache[T <: Clearable](cache: T): T = {
       cache match {
@@ -429,6 +417,22 @@ abstract class SymbolTable extends macros.Universe
       cache
     }
 
+    /** Closes the provided classloader at the conclusion of this Run */
+    final def recordClassloader(loader: ClassLoader): ClassLoader = {
+      def attemptClose(loader: ClassLoader): Unit = {
+        loader match {
+          case u: URLClassLoader => debuglog("Closing classloader " + u); u.close()
+          case _ =>
+        }
+      }
+      caches ::= new WeakReference((new Clearable {
+        def clear(): Unit = {
+          attemptClose(loader)
+        }
+      }))
+      loader
+    }
+
     /**
      * Removes a cache from the per-run caches. This is useful for testing: it allows running the
      * compiler and then inspect the state of a cache.
@@ -490,6 +494,7 @@ abstract class SymbolTable extends macros.Universe
     val changesBaseClasses = true
     def transform(sym: Symbol, tpe: Type): Type = tpe
   }
+  var nextFrom: Array[InfoTransformer] = null
 
   /** The phase which has given index as identifier. */
   val phaseWithId: Array[Phase]
