@@ -13,8 +13,9 @@
 package scala.tools.nsc
 
 import java.net.URL
-import util.Exceptional.unwrap
+import util.Exceptional.rootCause
 import scala.reflect.internal.util.ScalaClassLoader
+import scala.util.control.NonFatal
 
 trait CommonRunner {
   /** Run a given object, specified by name, using a
@@ -24,17 +25,15 @@ trait CommonRunner {
    *  @throws NoSuchMethodException
    *  @throws InvocationTargetException
    */
-  def run(urls: Seq[URL], objectName: String, arguments: Seq[String]) {
-    (ScalaClassLoader fromURLs urls).run(objectName, arguments)
-  }
+  def run(urls: Seq[URL], objectName: String, arguments: Seq[String]): Unit =
+    ScalaClassLoader.fromURLs(urls).run(objectName, arguments)
 
-  /** Catches exceptions enumerated by run (in the case of InvocationTargetException,
-   *  unwrapping it) and returns it any thrown in Left(x).
+  /** Catches any non-fatal exception thrown by run (in the case of InvocationTargetException,
+   *  unwrapping it) and returns it in an Option.
    */
-  def runAndCatch(urls: Seq[URL], objectName: String, arguments: Seq[String]): Either[Throwable, Boolean] = {
-    try   { run(urls, objectName, arguments) ; Right(true) }
-    catch { case e: Throwable => Left(unwrap(e)) }
-  }
+  def runAndCatch(urls: Seq[URL], objectName: String, arguments: Seq[String]): Option[Throwable] =
+    try   { run(urls, objectName, arguments) ; None }
+    catch { case NonFatal(e) => Some(rootCause(e)) }
 }
 
 /** An object that runs another object specified by name.
@@ -42,4 +41,4 @@ trait CommonRunner {
  *  @author  Lex Spoon
  *  @version 1.1, 2007/7/13
  */
-object ObjectRunner extends CommonRunner { }
+object ObjectRunner extends CommonRunner
