@@ -13,6 +13,7 @@
 package scala.tools.nsc
 package ast
 
+import scala.annotation.tailrec
 import symtab._
 import util.DocStrings._
 import scala.collection.mutable
@@ -34,7 +35,7 @@ trait DocComments { self: Global =>
    */
   val docComments = mutable.WeakHashMap[Symbol, DocComment]()
 
-  def clearDocComments() {
+  def clearDocComments(): Unit = {
     cookedDocComments.clear()
     docComments.clear()
     defs.clear()
@@ -61,7 +62,7 @@ trait DocComments { self: Global =>
     else sym.owner.ancestors map (sym overriddenSymbol _) filter (_ != NoSymbol)
   }
 
-  def fillDocComment(sym: Symbol, comment: DocComment) {
+  def fillDocComment(sym: Symbol, comment: DocComment): Unit = {
     docComments(sym) = comment
     comment.defineVariables(sym)
   }
@@ -261,8 +262,8 @@ trait DocComments { self: Global =>
               val sectionTextBounds = extractSectionText(parent, section)
               cleanupSectionText(parent.substring(sectionTextBounds._1, sectionTextBounds._2))
             case None =>
-              reporter.info(sym.pos, "The \"" + getSectionHeader + "\" annotation of the " + sym +
-                  " comment contains @inheritdoc, but the corresponding section in the parent is not defined.", force = true)
+              reporter.echo(sym.pos, "The \"" + getSectionHeader + "\" annotation of the " + sym +
+                  " comment contains @inheritdoc, but the corresponding section in the parent is not defined.")
               "<invalid inheritdoc annotation>"
           }
 
@@ -302,7 +303,8 @@ trait DocComments { self: Global =>
    *  @param vble  The variable for which a definition is searched
    *  @param site  The class for which doc comments are generated
    */
-  def lookupVariable(vble: String, site: Symbol): Option[String] = site match {
+  @tailrec
+  final def lookupVariable(vble: String, site: Symbol): Option[String] = site match {
     case NoSymbol => None
     case _        =>
       val searchList =
@@ -311,7 +313,8 @@ trait DocComments { self: Global =>
 
       searchList collectFirst { case x if defs(x) contains vble => defs(x)(vble) } match {
         case Some(str) if str startsWith "$" => lookupVariable(str.tail, site)
-        case res                             => res orElse lookupVariable(vble, site.owner)
+        case s @ Some(str)                   => s
+        case None                            => lookupVariable(vble, site.owner)
       }
   }
 
@@ -326,6 +329,7 @@ trait DocComments { self: Global =>
   protected def expandVariables(initialStr: String, sym: Symbol, site: Symbol): String = {
     val expandLimit = 10
 
+    @tailrec
     def expandInternal(str: String, depth: Int): String = {
       if (depth >= expandLimit)
         throw new ExpansionLimitExceeded(str)
@@ -341,7 +345,7 @@ trait DocComments { self: Global =>
         else {
           val vstart = idx
           idx = skipVariable(str, idx + 1)
-          def replaceWith(repl: String) {
+          def replaceWith(repl: String): Unit = {
             out append str.substring(copied, vstart)
             out append repl
             copied = idx
@@ -408,6 +412,8 @@ trait DocComments { self: Global =>
       val comment      = "/** " + raw.substring(commentStart, end) + "*/"
       val commentPos   = subPos(commentStart, end)
 
+      self.currentRun.reporting.deprecationWarning(codePos, "The @usecase tag is deprecated, instead use the @example tag to document the usage of your API", "2.13.0")
+
       UseCase(DocComment(comment, commentPos, codePos), code, codePos)
     }
 
@@ -488,7 +494,7 @@ trait DocComments { self: Global =>
           case _ =>
             (getSite(partnames.head), partnames.tail)
         }
-        val result = (start /: rest)(select(_, _, NoType))
+        val result = rest.foldLeft(start)(select(_, _, NoType))
         if (result == NoType)
           reporter.warning(comment.codePos, "Could not find the type " + variable + " points to while expanding it " +
                                             "for the usecase signature of " + sym + " in " + site + "." +
@@ -526,6 +532,7 @@ trait DocComments { self: Global =>
               (typeRef(NoPrefix, alias, Nil), false)
           }
 
+      @tailrec
       def subst(sym: Symbol, from: List[Symbol], to: List[(Type, Boolean)]): (Type, Boolean) =
         if (from.isEmpty) (sym.tpe, false)
         else if (from.head == sym) to.head
