@@ -1,19 +1,25 @@
-/* NSC -- new Scala compiler
- * Copyright 2005-2013 LAMP/EPFL
- * @author  Martin Odersky
+/*
+ * Scala (https://www.scala-lang.org)
+ *
+ * Copyright EPFL and Lightbend, Inc.
+ *
+ * Licensed under Apache License 2.0
+ * (http://www.apache.org/licenses/LICENSE-2.0).
+ *
+ * See the NOTICE file distributed with this work for
+ * additional information regarding copyright ownership.
  */
+
 package scala.tools.nsc
 package ast.parser
 
-import scala.tools.nsc.util.{ CharArrayReader, CharArrayReaderData }
+import scala.tools.nsc.util.{CharArrayReader, CharArrayReaderData}
 import scala.reflect.internal.util._
 import scala.reflect.internal.Chars._
 import Tokens._
-import scala.annotation.{ switch, tailrec }
-import scala.collection.mutable
-import mutable.{ ListBuffer, ArrayBuffer }
+import scala.annotation.{switch, tailrec}
+import scala.collection.mutable, mutable.{ListBuffer, ArrayBuffer}
 import scala.tools.nsc.ast.parser.xml.Utility.isNameStart
-import scala.language.postfixOps
 
 /** See Parsers.scala / ParsersCommon for some explanation of ScannersCommon.
  */
@@ -212,7 +218,7 @@ trait Scanners extends ScannersCommon {
 
     /** Clear buffer and set name and token */
     private def finishNamed(idtoken: Token = IDENTIFIER): Unit = {
-      name = newTermName(cbuf.toString)
+      name = newTermName(cbuf.toArray)
       cbuf.clear()
       token = idtoken
       if (idtoken == IDENTIFIER) {
@@ -433,6 +439,7 @@ trait Scanners extends ScannersCommon {
 
     /** read next token, filling TokenData fields of Scanner.
      */
+    @tailrec
     protected final def fetchToken(): Unit = {
       offset = charOffset - 1
       (ch: @switch) match {
@@ -543,24 +550,49 @@ trait Scanners extends ScannersCommon {
           }
           fetchDoubleQuote()
         case '\'' =>
+          def unclosedCharLit() = {
+            val unclosed = "unclosed character literal"
+            // advise if previous token was Symbol contiguous with the orphan single quote at offset
+            val msg = {
+              val maybeMistakenQuote =
+                this match {
+                  case sfs: SourceFileScanner =>
+                    val wholeLine = sfs.source.lineToString(sfs.source.offsetToLine(offset))
+                    wholeLine.count(_ == '\'') > 1
+                  case _ => false
+                }
+              if (token == SYMBOLLIT && offset == lastOffset) s"""$unclosed (or use " for string literal "$strVal")"""
+              else if (maybeMistakenQuote) s"""$unclosed (or use " not ' for string literal)"""
+              else unclosed
+            }
+            syntaxError(msg)
+          }
           def fetchSingleQuote() = {
             nextChar()
             if (isIdentifierStart(ch))
-              charLitOr(getIdentRest)
+              charLitOr(() => getIdentRest())
             else if (isOperatorPart(ch) && (ch != '\\'))
-              charLitOr(getOperatorRest)
+              charLitOr(() => getOperatorRest())
             else if (!isAtEnd && (ch != SU && ch != CR && ch != LF || isUnicodeEscape)) {
+              val isEmptyCharLit = (ch == '\'')
               getLitChar()
               if (ch == '\'') {
-                nextChar()
-                token = CHARLIT
-                setStrVal()
+                if (isEmptyCharLit && currentRun.isScala213)
+                  syntaxError("empty character literal (use '\\'' for single quote)")
+                else {
+                  if (isEmptyCharLit)
+                    deprecationWarning("deprecated syntax for character literal (use '\\'' for single quote)", "2.12.2")
+                  nextChar()
+                  token = CHARLIT
+                  setStrVal()
+                }
+              } else if (isEmptyCharLit) {
+                syntaxError("empty character literal")
               } else {
-                syntaxError("unclosed character literal")
+                unclosedCharLit()
               }
             }
-            else
-              syntaxError("unclosed character literal")
+            else unclosedCharLit()
           }
           fetchSingleQuote()
         case '.' =>
@@ -596,8 +628,10 @@ trait Scanners extends ScannersCommon {
         case _ =>
           def fetchOther() = {
             if (ch == '\u21D2') {
+              deprecationWarning("The unicode arrow `⇒` is deprecated, use `=>` instead. If you still wish to display it as one character, consider using a font with programming ligatures such as Fira Code.", "2.13.0")
               nextChar(); token = ARROW
             } else if (ch == '\u2190') {
+              deprecationWarning("The unicode arrow `←` is deprecated, use `<-` instead. If you still wish to display it as one character, consider using a font with programming ligatures such as Fira Code.", "2.13.0")
               nextChar(); token = LARROW
             } else if (Character.isUnicodeIdentifierStart(ch)) {
               putChar(ch)
@@ -649,6 +683,7 @@ trait Scanners extends ScannersCommon {
       else syntaxError("unclosed quoted identifier")
     }
 
+    @tailrec
     private def getIdentRest(): Unit = (ch: @switch) match {
       case 'A' | 'B' | 'C' | 'D' | 'E' |
            'F' | 'G' | 'H' | 'I' | 'J' |
@@ -683,6 +718,7 @@ trait Scanners extends ScannersCommon {
         }
     }
 
+    @tailrec
     private def getOperatorRest(): Unit = (ch: @switch) match {
       case '~' | '!' | '@' | '#' | '%' |
            '^' | '*' | '+' | '-' | '<' |
@@ -792,7 +828,7 @@ trait Scanners extends ScannersCommon {
             next.token = kwArray(idx)
           }
         } else {
-          syntaxError("invalid string interpolation: `$$', `$'ident or `$'BlockExpr expected")
+          syntaxError(s"invalid string interpolation $$$ch, expected: $$$$, $$identifier or $${expression}")
         }
       } else {
         val isUnclosedLiteral = !isUnicodeEscape && (ch == SU || (!multiLine && (ch == CR || ch == LF)))
@@ -855,11 +891,7 @@ trait Scanners extends ScannersCommon {
             }
           }
           val alt = if (oct == LF) "\\n" else "\\u%04x" format oct
-          def msg(what: String) = s"Octal escape literals are $what, use $alt instead."
-          if (settings.future)
-            syntaxError(start, msg("unsupported"))
-          else
-            deprecationWarning(start, msg("deprecated"), "2.11.0")
+          syntaxError(start, s"octal escape literals are unsupported: use $alt instead")
           putChar(oct.toChar)
         } else {
           ch match {
@@ -894,11 +926,11 @@ trait Scanners extends ScannersCommon {
      *  if one is present.
      */
     protected def getFraction(): Unit = {
-      token = DOUBLELIT
-      while ('0' <= ch && ch <= '9') {
+      while ('0' <= ch && ch <= '9' || isNumberSeparator(ch)) {
         putChar(ch)
         nextChar()
       }
+      checkNoTrailingSeparator()
       if (ch == 'e' || ch == 'E') {
         val lookahead = lookaheadReader
         lookahead.nextChar()
@@ -912,10 +944,11 @@ trait Scanners extends ScannersCommon {
             putChar(ch)
             nextChar()
           }
-          while ('0' <= ch && ch <= '9') {
+          while ('0' <= ch && ch <= '9' || isNumberSeparator(ch)) {
             putChar(ch)
             nextChar()
           }
+          checkNoTrailingSeparator()
         }
         token = DOUBLELIT
       }
@@ -927,7 +960,8 @@ trait Scanners extends ScannersCommon {
         putChar(ch)
         nextChar()
         token = FLOATLIT
-      }
+      } else
+        token = DOUBLELIT
       checkNoLetter()
       setStrVal()
     }
@@ -943,39 +977,41 @@ trait Scanners extends ScannersCommon {
      *  path, attempts to write base 8 literals except `0` emit a verbose error.
      */
     def intVal(negated: Boolean): Long = {
-      def malformed: Long = {
-        if (base == 8) syntaxError("Decimal integer literals may not have a leading zero. (Octal syntax is obsolete.)")
-        else syntaxError("malformed integer number")
-        0
-      }
-      def tooBig: Long = {
-        syntaxError("integer number too large")
-        0
-      }
       def intConvert: Long = {
+        def malformed: Long = { syntaxError("malformed integer number") ; 0 }
+        def tooBig: Long = { syntaxError("integer number too large") ; 0 }
         val len = strVal.length
         if (len == 0) {
           if (base != 8) syntaxError("missing integer number")  // e.g., 0x;
-          0
+          0                                                     // 0 still looks like octal prefix
         } else {
+          if (base == 8) {
+            if (settings.warnOctalLiteral)
+              deprecationWarning("Decimal integer literals should not have a leading zero. (Octal syntax is obsolete.)" , since="2.10")
+            base = 10
+          }
           val divider     = if (base == 10) 1 else 2
           val limit: Long = if (token == LONGLIT) Long.MaxValue else Int.MaxValue
           @tailrec def convert(value: Long, i: Int): Long =
             if (i >= len) value
             else {
-              val d = digit2int(strVal charAt i, base)
-              if (d < 0)
-                malformed
-              else if (value < 0 ||
-                  limit / (base / divider) < value ||
-                  limit - (d / divider) < value * (base / divider) &&
-                  !(negated && limit == value * base - 1 + d))
-                tooBig
-              else
-                convert(value * base + d, i + 1)
+              val c = strVal.charAt(i)
+              if (isNumberSeparator(c)) convert(value, i + 1)
+              else {
+                val d = digit2int(c, base)
+                if (d < 0)
+                  malformed
+                else if (value < 0 ||
+                    limit / (base / divider) < value ||
+                    limit - (d / divider) < value * (base / divider) &&
+                    !(negated && limit == value * base - 1 + d))
+                  tooBig
+                else
+                  convert(value * base + d, i + 1)
+              }
             }
           val result = convert(0, 0)
-          if (base == 8) malformed else if (negated) -result else result
+          if (negated) -result else result
         }
       }
       if (token == CHARLIT && !negated) charVal.toLong else intConvert
@@ -988,11 +1024,12 @@ trait Scanners extends ScannersCommon {
     /** Convert current strVal, base to float value.
      */
     def floatVal(negated: Boolean): Float = {
+      val text = removeNumberSeparators(strVal)
       try {
-        val value: Float = java.lang.Float.parseFloat(strVal)
+        val value: Float = java.lang.Float.parseFloat(text)
         if (value > Float.MaxValue)
           syntaxError("floating point number too large")
-        if (value == 0.0f && !zeroFloat.pattern.matcher(strVal).matches)
+        if (value == 0.0f && !zeroFloat.pattern.matcher(text).matches)
           syntaxError("floating point number too small")
         if (negated) -value else value
       } catch {
@@ -1007,11 +1044,12 @@ trait Scanners extends ScannersCommon {
     /** Convert current strVal, base to double value.
      */
     def doubleVal(negated: Boolean): Double = {
+      val text = removeNumberSeparators(strVal)
       try {
-        val value: Double = java.lang.Double.parseDouble(strVal)
+        val value: Double = java.lang.Double.parseDouble(text)
         if (value > Double.MaxValue)
           syntaxError("double precision floating point number too large")
-        if (value == 0.0d && !zeroFloat.pattern.matcher(strVal).matches)
+        if (value == 0.0d && !zeroFloat.pattern.matcher(text).matches)
           syntaxError("double precision floating point number too small")
         if (negated) -value else value
       } catch {
@@ -1028,6 +1066,18 @@ trait Scanners extends ScannersCommon {
         syntaxError("Invalid literal number")
     }
 
+    @inline private def isNumberSeparator(c: Char): Boolean = c == '_' //|| c == '\''
+
+    @inline private def removeNumberSeparators(s: String): String =
+      if (s.indexOf('_') > 0) s.replaceAllLiterally("_", "") /*.replaceAll("'","")*/ else s
+
+    // disallow trailing numeric separator char, but let lexing limp along
+    def checkNoTrailingSeparator(): Unit =
+      if (cbuf.nonEmpty && isNumberSeparator(cbuf.last)) {
+        syntaxError(offset + cbuf.length - 1, "trailing separator is not allowed")
+        cbuf.setLength(cbuf.length - 1)
+      }
+
     /** Read a number into strVal.
      *
      *  The `base` can be 8, 10 or 16, where base 8 flags a leading zero.
@@ -1036,22 +1086,27 @@ trait Scanners extends ScannersCommon {
     protected def getNumber(): Unit = {
       // consume digits of a radix
       def consumeDigits(radix: Int): Unit =
-        while (digit2int(ch, radix) >= 0) {
+        while (isNumberSeparator(ch) || digit2int(ch, radix) >= 0) {
           putChar(ch)
           nextChar()
         }
-      // adding decimal point is always OK because `Double valueOf "0."` is OK
+      // at dot with digit following
       def restOfNonIntegralNumber(): Unit = {
         putChar('.')
-        if (ch == '.') nextChar()
+        nextChar()
         getFraction()
       }
+      // 1l is an acknowledged bad practice
+      def lintel(): Unit = {
+        val msg = "Lowercase el for long is not recommended because it is easy to confuse with numeral 1; use uppercase L instead"
+        if (ch == 'l') deprecationWarning(offset + cbuf.length, msg, since="2.13.0")
+      }
       // after int: 5e7f, 42L, 42.toDouble but not 42b. Repair 0d.
       def restOfNumber(): Unit = {
         ch match {
           case 'e' | 'E' | 'f' | 'F' |
-               'd' | 'D' => if (cbuf.isEmpty) putChar('0'); restOfNonIntegralNumber()
-          case 'l' | 'L' => token = LONGLIT ; setStrVal() ; nextChar()
+               'd' | 'D' => if (cbuf.isEmpty) putChar('0'); getFraction()
+          case 'l' | 'L' => lintel() ; token = LONGLIT ; setStrVal() ; nextChar()
           case _         => token = INTLIT  ; setStrVal() ; checkNoLetter()
         }
       }
@@ -1059,6 +1114,8 @@ trait Scanners extends ScannersCommon {
       // consume leading digits, provisionally an Int
       consumeDigits(if (base == 16) 16 else 10)
 
+      checkNoTrailingSeparator()
+
       val detectedFloat: Boolean = base != 16 && ch == '.' && isDigit(lookaheadReader.getc)
       if (detectedFloat) restOfNonIntegralNumber() else restOfNumber()
     }
@@ -1306,7 +1363,7 @@ trait Scanners extends ScannersCommon {
 //        println("applying brace patch "+offset)//DEBUG
         if (patch.inserted) {
           next copyFrom this
-          error(offset, "Missing closing brace `}' assumed here")
+          error(offset, "Missing closing brace `}` assumed here")
           token = RBRACE
           true
         } else {
@@ -1324,7 +1381,7 @@ trait Scanners extends ScannersCommon {
     /** The source code with braces and line starts annotated with [NN] showing the index */
     private def markedSource = {
       val code   = unit.source.content
-      val braces = code.indices filter (idx => "{}\n" contains code(idx)) toSet;
+      val braces = code.indices.filter(idx => "{}\n" contains code(idx)).toSet
       val mapped = code.indices map (idx => if (braces(idx)) s"${code(idx)}[$idx]" else "" + code(idx))
       mapped.mkString("")
     }
@@ -1343,7 +1400,7 @@ trait Scanners extends ScannersCommon {
       var lineCount = 1
       var lastOffset = 0
       var indent = 0
-      val oldBalance = scala.collection.mutable.Map[Int, Int]()
+      val oldBalance = mutable.Map[Int, Int]()
       def markBalance() = for ((k, v) <- balance) oldBalance(k) = v
       markBalance()
 
@@ -1430,6 +1487,7 @@ trait Scanners extends ScannersCommon {
     var tabSeen = false
 
     def line(offset: Offset): Int = {
+      @tailrec
       def findLine(lo: Int, hi: Int): Int = {
         val mid = (lo + hi) / 2
         if (offset < lineStart(mid)) findLine(lo, mid - 1)
