@@ -16,19 +16,19 @@ package scala.tools
 package nsc
 package settings
 
-import io.{ AbstractFile, Jar, Path, PlainFile, VirtualDirectory }
-import scala.collection.generic.Clearable
+import io.{AbstractFile, Jar, Path, PlainFile, VirtualDirectory}
+import scala.annotation.tailrec
+import scala.collection.mutable.Clearable
 import scala.io.Source
-import scala.reflect.internal.util.{ SomeOfNil, StringOps }
-import scala.reflect.{ ClassTag, classTag }
+import scala.reflect.internal.util.{SomeOfNil, StringOps}
+import scala.reflect.{ClassTag, classTag}
 
 /** A mutable Settings object.
  */
 class MutableSettings(val errorFn: String => Unit)
               extends scala.reflect.internal.settings.MutableSettings
                  with AbsSettings
-                 with ScalaSettings
-                 with Mutable {
+                 with ScalaSettings {
   type ResultOfTryToSet = List[String]
 
   def withErrorFn(errorFn: String => Unit): MutableSettings = {
@@ -37,7 +37,7 @@ class MutableSettings(val errorFn: String => Unit)
     settings
   }
 
-  def copyInto(settings: MutableSettings) {
+  def copyInto(settings: MutableSettings): Unit = {
     allSettings foreach { thisSetting =>
       val otherSetting = settings.allSettings find { _.name == thisSetting.name }
       otherSetting foreach { otherSetting =>
@@ -66,6 +66,7 @@ class MutableSettings(val errorFn: String => Unit)
    *  Returns (success, List of unprocessed arguments)
    */
   def processArguments(arguments: List[String], processAll: Boolean): (Boolean, List[String]) = {
+    @tailrec
     def loop(args: List[String], residualArgs: List[String]): (Boolean, List[String]) = args match {
       case Nil        =>
         (checkDependencies, residualArgs)
@@ -147,49 +148,35 @@ class MutableSettings(val errorFn: String => Unit)
       if (s endsWith ":") {
         clearIfExists(s.init)
       } else {
-        for {
-          (p, args) <- StringOps.splitWhere(s, _ == ':', doDropIndex = true)
-          rest      <- tryToSetIfExists(p, (args split ",").toList, (s: Setting) => s.tryToSetColon _)
-        } yield rest
+        StringOps.splitWhere(s, _ == ':', doDropIndex = true).flatMap {
+          case (p, args) =>
+            tryToSetIfExists(p, (args split ",").toList, (s: Setting) => s.tryToSetColon(_))
+        }
       }
 
     // if arg is of form -Xfoo or -Xfoo bar (name = "-Xfoo")
     def parseNormalArg(p: String, args: List[String]): Option[List[String]] =
-      tryToSetIfExists(p, args, (s: Setting) => s.tryToSet _)
+      tryToSetIfExists(p, args, (s: Setting) => s.tryToSet(_))
 
     args match {
       case Nil          => Nil
+      case "-" :: rest  => errorFn("'-' is not a valid argument.") ; args
+      case arg :: rest if !arg.startsWith("-") => errorFn(s"Argument '$arg' does not start with '-'.") ; args
       case arg :: rest  =>
-        if (!arg.startsWith("-")) {
-          errorFn("Argument '" + arg + "' does not start with '-'.")
-          args
-        }
-        else if (arg == "-") {
-          errorFn("'-' is not a valid argument.")
-          args
-        }
-        else {
-          // we dispatch differently based on the appearance of p:
-          // 1) If it matches a prefix setting it is sent there directly.
-          // 2) If it has a : it is presumed to be -Xfoo:bar,baz
-          // 3) Otherwise, the whole string should be a command name
-          //
-          // Internally we use Option[List[String]] to discover error,
-          // but the outside expects our arguments back unchanged on failure
-          val prefix = prefixSettings find (_ respondsTo arg)
-          if (prefix.isDefined) {
-            prefix.get tryToSet args
-            rest
-          }
-          else if (arg contains ":") parseColonArg(arg) match {
-            case Some(_)  => rest
-            case None     => args
-          }
-          else parseNormalArg(arg, rest) match {
-            case Some(xs) => xs
-            case None     => args
-          }
+        // we dispatch differently based on the appearance of p:
+        // 1) If it matches a prefix setting it is sent there directly.
+        // 2) If it has a : it is presumed to be -Xfoo:bar,baz
+        // 3) Otherwise, the whole string should be a command name
+        //
+        // Internally we use Option[List[String]] to discover error,
+        // but the outside expects our arguments back unchanged on failure
+        val prefix = prefixSettings.find(_ respondsTo arg)
+        prefix.map { setting => setting.tryToSet(args); rest }
+        .orElse {
+          if (arg contains ":") parseColonArg(arg).map(_ => rest)
+          else parseNormalArg(arg, rest)
         }
+        .getOrElse(args)
     }
   }
 
@@ -204,7 +191,7 @@ class MutableSettings(val errorFn: String => Unit)
   * The class loader for `T` should provide resources `app.class.path`
   * and `boot.class.path`.  These resources should contain the application
   * and boot classpaths in the same form as would be passed on the command line.*/
-  def embeddedDefaults(loader: ClassLoader) {
+  def embeddedDefaults(loader: ClassLoader): Unit = {
     explicitParentLoader = Option(loader) // for the Interpreter parentClassLoader
     getClasspath("app", loader) foreach { classpath.value = _ }
     getClasspath("boot", loader) foreach { bootclasspath append _ }
@@ -216,9 +203,10 @@ class MutableSettings(val errorFn: String => Unit)
   /** Retrieves the contents of resource "${id}.class.path" from `loader`
   * (wrapped in Some) or None if the resource does not exist.*/
   private def getClasspath(id: String, loader: ClassLoader): Option[String] =
-    Option(loader).flatMap(ld => Option(ld.getResource(id + ".class.path"))).map { cp =>
-       Source.fromURL(cp).mkString
-    }
+    for {
+      ld <- Option(loader)
+      r  <- Option(ld.getResource(s"$id.class.path"))
+    } yield Source.fromURL(r).mkString
 
   // a wrapper for all Setting creators to keep our list up to date
   private def add[T <: Setting](s: T): T = {
@@ -289,7 +277,7 @@ class MutableSettings(val errorFn: String => Unit)
     /** Set the single output directory. From now on, all files will
      *  be dumped in there, regardless of previous calls to 'add'.
      */
-    def setSingleOutput(outDir: String) {
+    def setSingleOutput(outDir: String): Unit = {
       val dst = AbstractFile.getDirectory(outDir)
       setSingleOutput(checkDir(dst, outDir, allowJar = true))
     }
@@ -299,11 +287,11 @@ class MutableSettings(val errorFn: String => Unit)
     /** Set the single output directory. From now on, all files will
      *  be dumped in there, regardless of previous calls to 'add'.
      */
-    def setSingleOutput(dir: AbstractFile) {
+    def setSingleOutput(dir: AbstractFile): Unit = {
       singleOutDir = Some(dir)
     }
 
-    def add(src: AbstractFile, dst: AbstractFile) {
+    def add(src: AbstractFile, dst: AbstractFile): Unit = {
       singleOutDir = None
       outputDirs ::= ((src, dst))
     }
@@ -365,19 +353,19 @@ class MutableSettings(val errorFn: String => Unit)
   /** A base class for settings of all types.
    *  Subclasses each define a `value` field of the appropriate type.
    */
-  abstract class Setting(val name: String, val helpDescription: String) extends AbsSetting with SettingValue with Mutable {
+  abstract class Setting(val name: String, val helpDescription: String) extends AbsSetting with SettingValue {
     /** Will be called after this Setting is set for any extra work. */
-    private var _postSetHook: this.type => Unit = (x: this.type) => ()
+    private[this] var _postSetHook: this.type => Unit = (x: this.type) => ()
     override def postSetHook(): Unit = _postSetHook(this)
     def withPostSetHook(f: this.type => Unit): this.type = { _postSetHook = f ; this }
 
     /** The syntax defining this setting in a help string */
-    private var _helpSyntax = name
+    private[this] var _helpSyntax = name
     override def helpSyntax: String = _helpSyntax
     def withHelpSyntax(s: String): this.type    = { _helpSyntax = s ; this }
 
     /** Abbreviations for this setting */
-    private var _abbreviations: List[String] = Nil
+    private[this] var _abbreviations: List[String] = Nil
     override def abbreviations = _abbreviations
     def withAbbreviation(s: String): this.type  = { _abbreviations ++= List(s) ; this }
 
@@ -386,7 +374,7 @@ class MutableSettings(val errorFn: String => Unit)
     override def dependencies = dependency.toList
     def dependsOn(s: Setting, value: String): this.type = { dependency = Some((s, value)); this }
 
-    private var _deprecationMessage: Option[String] = None
+    private[this] var _deprecationMessage: Option[String] = None
     override def deprecationMessage = _deprecationMessage
     def withDeprecationMessage(msg: String): this.type = { _deprecationMessage = Some(msg) ; this }
   }
@@ -429,12 +417,7 @@ class MutableSettings(val errorFn: String => Unit)
     // Ensure that the default value is actually valid
     assert(isInputValid(default))
 
-    def parseArgument(x: String): Option[Int] = {
-      parser(x) orElse {
-        try   { Some(x.toInt) }
-        catch { case _: NumberFormatException => None }
-      }
-    }
+    def parseArgument(x: String): Option[Int] = parser(x) orElse x.toIntOption
 
     def errorMsg() = errorFn(s"invalid setting for $name $getValidText")
 
@@ -463,7 +446,7 @@ class MutableSettings(val errorFn: String => Unit)
 
     def tryToSet(args: List[String]) = { value = true ; Some(args) }
     def unparse: List[String] = if (value) List(name) else Nil
-    override def tryToSetFromPropertyValue(s : String) { // used from ide
+    override def tryToSetFromPropertyValue(s : String): Unit = { // used from ide
       value = s.equalsIgnoreCase("true")
     }
     override def tryToSetColon(args: List[String]) = args match {
@@ -589,7 +572,7 @@ class MutableSettings(val errorFn: String => Unit)
     default: String)
     extends StringSetting("-d", "directory|jar", "destination for generated classfiles.", default, None) {
       value = default
-      override def value_=(str: String) {
+      override def value_=(str: String): Unit = {
         super.value_=(str)
         try outputDirs.setSingleOutput(str)
         catch { case FatalError(msg) => errorFn(msg) }
@@ -643,9 +626,9 @@ class MutableSettings(val errorFn: String => Unit)
     descr: String,
     val domain: E,
     val default: Option[List[String]]
-  ) extends Setting(name, s"$descr: `_' for all, `$name:help' to list choices.") with Clearable {
+  ) extends Setting(name, descr) with Clearable {
 
-    withHelpSyntax(s"$name:<_,$helpArg,-$helpArg>")
+    withHelpSyntax(s"$name:<${helpArg}s>")
 
     object ChoiceOrVal {
       def unapply(a: domain.Value): Option[(String, String, List[domain.Choice])] = a match {
@@ -744,6 +727,7 @@ class MutableSettings(val errorFn: String => Unit)
         case s if isChoice(s) => added += s // this case also adds "_"
         case s                => badChoice(s)
       }
+      @tailrec
       def loop(args: List[String]): List[String] = args match {
         case arg :: _ if halting && (!isPos(arg) || !isChoice(arg)) => args
         case arg :: rest => tryArg(arg) ; loop(rest)
@@ -816,6 +800,7 @@ class MutableSettings(val errorFn: String => Unit)
 
     // try to set. halting means halt at first non-arg
     protected def tryToSetArgs(args: List[String], halting: Boolean) = {
+      @tailrec
       def loop(args: List[String]): List[String] = args match {
         case arg :: rest =>
           if (halting && (arg startsWith "-")) args
@@ -852,8 +837,8 @@ class MutableSettings(val errorFn: String => Unit)
     val default: String,
     val choicesHelp: List[String])
   extends Setting(name,
-    if (choicesHelp.isEmpty) s"$descr Choices: ${choices.mkString("(", ",", ")")}, default: $default."
-    else s"$descr Default: `$default', `help' to list choices.") {
+    if (choicesHelp.isEmpty) s"$descr ${choices.map(s => if (s == default) s"[$s]" else s).mkString("(", ",", ")")}"
+    else s"$descr Default: `$default`, `help` to list choices.") {
     type T = String
     protected var v: T = default
     def indexOfChoice: Int = choices indexOf value
@@ -894,9 +879,8 @@ class MutableSettings(val errorFn: String => Unit)
   }
 
   /** A setting represented by a list of strings which should be prefixes of
-   *  phase names. This is not checked here, however.  Alternatively the string
-   *  `"all"` can be used to represent all phases.
-   *  (the empty list, unless set)
+   *  phase names. This is not checked here, however.  Alternatively, underscore
+   *  can be used to indicate all phases.
    */
   class PhasesSetting private[nsc](
     name: String,
@@ -924,7 +908,7 @@ class MutableSettings(val errorFn: String => Unit)
       _names = numsAndStrs._2
       _v     = t
     }
-    override def value = if (v contains "all") List("all") else super.value // i.e., v
+    override def value = if (v contains "_") List("_") else super.value // i.e., v
     private def numericValues = _numbs
     private def stringValues  = _names
     private def phaseIdTest(i: Int): Boolean = numericValues exists (_ match {
@@ -948,14 +932,14 @@ class MutableSettings(val errorFn: String => Unit)
     def clear(): Unit = (v = Nil)
 
     // we slightly abuse the usual meaning of "contains" here by returning
-    // true if our phase list contains "all", regardless of the incoming argument
+    // true if our phase list contains "_", regardless of the incoming argument
     def contains(phName: String)     = doAllPhases || containsName(phName)
     def containsName(phName: String) = stringValues exists (phName startsWith _)
     def containsId(phaseId: Int)     = phaseIdTest(phaseId)
     def containsPhase(ph: Phase)     = contains(ph.name) || containsId(ph.id)
 
-    def doAllPhases = stringValues contains "all"
-    def unparse: List[String] = value map (name + ":" + _)
+    def doAllPhases = stringValues.contains("_")
+    def unparse: List[String] = value.map(v => s"$name:$v")
 
     withHelpSyntax(
       if (default == "") name + ":<phases>"
