@@ -13,14 +13,14 @@
 package scala
 package tools.nsc.transform.patmat
 
-import scala.language.postfixOps
 import scala.collection.mutable
+import scala.collection.immutable.ArraySeq
 import scala.reflect.internal.util.{HashSet, NoPosition, Position, StatisticsStatics}
 
 trait Logic extends Debugging  {
   import global.statistics
 
-  private def max(xs: Seq[Int]) = if (xs isEmpty) 0 else xs max
+  private def max(xs: Seq[Int]) = if (xs.isEmpty) 0 else xs.max
   private def alignedColumns(cols: Seq[Any]): Seq[String] = {
     def toString(x: Any) = if (x == null) "" else x.toString
     if (cols.isEmpty || cols.tails.isEmpty) cols map toString
@@ -156,7 +156,7 @@ trait Logic extends Debugging  {
         val newSym = new Sym(variable, const)
         (uniques findEntryOrUpdate newSym)
       }
-      def nextSymId = {_symId += 1; _symId}; private var _symId = 0
+      def nextSymId = {_symId += 1; _symId}; private[this] var _symId = 0
       implicit val SymOrdering: Ordering[Sym] = Ordering.by(_.id)
     }
 
@@ -229,10 +229,20 @@ trait Logic extends Debugging  {
         }
       }
 
+      def mapConserve[A <: AnyRef](s: Set[A])(f: A => A): Set[A] = {
+        var changed = false
+        val s1 = s.map {a =>
+          val a1 = f(a)
+          if (a1 ne a) changed = true
+          a1
+        }
+        if (changed) s1 else s
+      }
+
       // push negation inside formula
       def negationNormalFormNot(p: Prop): Prop = p match {
-        case And(ops) => Or(ops.map(negationNormalFormNot)) // De Morgan
-        case Or(ops)  => And(ops.map(negationNormalFormNot)) // De Morgan
+        case And(ops) => Or(mapConserve(ops)(negationNormalFormNot)) // De Morgan
+        case Or(ops)  => And(mapConserve(ops)(negationNormalFormNot)) // De Morgan
         case Not(p)   => negationNormalForm(p)
         case True     => False
         case False    => True
@@ -240,8 +250,12 @@ trait Logic extends Debugging  {
       }
 
       def negationNormalForm(p: Prop): Prop = p match {
-        case And(ops)     => And(ops.map(negationNormalForm))
-        case Or(ops)      => Or(ops.map(negationNormalForm))
+        case And(ops)     =>
+          val ops1 = mapConserve(ops)(negationNormalForm)
+          if (ops1 eq ops) p else And(ops1)
+        case Or(ops)      =>
+          val ops1 = mapConserve(ops)(negationNormalForm)
+          if (ops1 eq ops) p else Or(ops1)
         case Not(negated) => negationNormalFormNot(negated)
         case True
              | False
@@ -325,20 +339,20 @@ trait Logic extends Debugging  {
       def applySymbol(x: Sym): Unit = {}
     }
 
-    def gatherVariables(p: Prop): Set[Var] = {
+    def gatherVariables(p: Prop): collection.Set[Var] = {
       val vars = new mutable.HashSet[Var]()
       (new PropTraverser {
         override def applyVar(v: Var) = vars += v
       })(p)
-      vars.toSet
+      vars
     }
 
-    def gatherSymbols(p: Prop): Set[Sym] = {
+    def gatherSymbols(p: Prop): collection.Set[Sym] = {
       val syms = new mutable.HashSet[Sym]()
       (new PropTraverser {
         override def applySymbol(s: Sym) = syms += s
       })(p)
-      syms.toSet
+      syms
     }
 
     trait PropMap {
@@ -367,10 +381,8 @@ trait Logic extends Debugging  {
 
     }
 
-    // TODO: remove since deprecated
-    val budgetProp = scala.sys.Prop[String]("scalac.patmat.analysisBudget")
-    if (budgetProp.isSet) {
-      reportWarning(s"Please remove -D${budgetProp.key}, it is ignored.")
+    if (System.getProperty("scalac.patmat.analysisBudget") != null) {
+      reportWarning(s"Please remove -Dscalac.patmat.analysisBudget, it is ignored.")
     }
 
     // convert finite domain propositional logic with subtyping to pure boolean propositional logic
@@ -417,7 +429,7 @@ trait Logic extends Debugging  {
 
       val pure = props map (p => rewriteEqualsToProp(p))
 
-      val eqAxioms = mutable.ArrayBuffer[Prop]()
+      val eqAxioms = ArraySeq.newBuilder[Prop]
       @inline def addAxiom(p: Prop) = eqAxioms += p
 
       debug.patmat("removeVarEq vars: "+ vars)
@@ -459,12 +471,13 @@ trait Logic extends Debugging  {
         }
       }
 
-      debug.patmat(s"eqAxioms:\n${eqAxioms.mkString("\n")}")
+      val eqAxiomsSeq = eqAxioms.result()
+      debug.patmat(s"eqAxioms:\n${eqAxiomsSeq.mkString("\n")}")
       debug.patmat(s"pure:${pure.mkString("\n")}")
 
       if (StatisticsStatics.areSomeColdStatsEnabled) statistics.stopTimer(statistics.patmatAnaVarEq, start)
 
-      (And(eqAxioms: _*), pure)
+      (And(eqAxiomsSeq: _*), pure)
     }
 
     type Solvable
@@ -498,7 +511,7 @@ trait ScalaLogic extends Interface with Logic with TreeAndTypeAnalysis {
     def prepareNewAnalysis(): Unit = { Var.resetUniques(); Const.resetUniques() }
 
     object Var extends VarExtractor {
-      private var _nextId = 0
+      private[this] var _nextId = 0
       def nextId = {_nextId += 1; _nextId}
 
       def resetUniques() = {_nextId = 0; uniques.clear()}
@@ -714,10 +727,10 @@ trait ScalaLogic extends Interface with Logic with TreeAndTypeAnalysis {
     object Const {
       def resetUniques() = {_nextTypeId = 0; _nextValueId = 0; uniques.clear() ; trees.clear()}
 
-      private var _nextTypeId = 0
+      private[this] var _nextTypeId = 0
       def nextTypeId = {_nextTypeId += 1; _nextTypeId}
 
-      private var _nextValueId = 0
+      private[this] var _nextValueId = 0
       def nextValueId = {_nextValueId += 1; _nextValueId}
 
       private val uniques = new mutable.HashMap[Type, Const]
@@ -782,7 +795,7 @@ trait ScalaLogic extends Interface with Logic with TreeAndTypeAnalysis {
     // (At least conceptually: `true` is an instance of class `Boolean`)
     private def widenToClass(tp: Type): Type =
       if (tp.typeSymbol.isClass) tp
-      else if (tp.baseClasses.isEmpty) sys.error("Bad type: " + tp)
+      else if (tp.baseClasses.isEmpty) throw new IllegalArgumentException("Bad type: " + tp)
       else tp.baseType(tp.baseClasses.head)
 
     object TypeConst extends TypeConstExtractor {
