@@ -37,7 +37,7 @@ trait FrontEnd {
   val infos = new scala.collection.mutable.LinkedHashSet[Info]
 
   /** Handles incoming info */
-  def log(pos: Position, msg: String, severity: Severity) {
+  def log(pos: Position, msg: String, severity: Severity): Unit = {
     infos += Info(pos, msg, severity)
     severity.count += 1
     display(infos.last)
